@@ -7,7 +7,7 @@
 ** the Free Software Foundation; either version 2 of the License, or
 ** (at your option) any later version.
 **
-** This program is dsitributed in the hope that it will be useful,
+** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILIY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ** General Public License for more details.
@@ -20,6 +20,8 @@
 ** jwhitham@globalnet.co.uk
 */
 
+/* htget version 0.91 */
+
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
@@ -51,45 +53,34 @@ const char base64_table[64]= "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
 #define OPTIONS			15
 
 /* The HTGet function can return these errors: */
-#define ERROR_TIMEOUT		1
-#define ERROR_WRITE		2
-#define ERROR_ABORT		3
+#define ERROR_TIMEOUT			1
+#define ERROR_WRITE				2
+#define ERROR_ABORT				3
 #define ERROR_RESUME_IMPOSSIBLE	4
-#define ERROR_HTTP		5
-#define ERROR_NOT_NEEDED	6	/* file is already complete */
-#define ERROR_TCP		7
-#define DOWNLOAD_OK		8
-#define ERROR_PARSE		9
-#define ERROR_LIST		10	
+#define ERROR_HTTP				5
+#define ERROR_NOT_NEEDED		6	/* file is already complete */
+#define ERROR_TCP				7
+#define DOWNLOAD_OK				8
+#define ERROR_PARSE				9
+#define ERROR_LIST				10	
+#define ERROR_FOUR_OH_FOUR		11	
 
-#define BANNER	"HTGET (c) J Whitham 1998  <jwhitham@globalnet.co.uk> version 0.1\n" \
-		"See README. This program comes with NO WARRANTY of any kind.\n\n" 
 	
-const char	LongOption [ OPTIONS ][ 16 ] = { "--noresume" ,
-		"--notimeout" , "--timeout" , "--redial" ,
-		"--silent" , "--nocheck" , "--list" , "--logfile" ,
-		"--directory" , "--noestimate" , "--minrate" ,
-		"--stdout" , "--verbose" , "--be" , "--help" } ;
-const char	ShortOption [ OPTIONS ][ 4 ] = { "-nr" ,
-		"-nt" , "-t" , "-r" ,
-		"-s" , "-nc" , "-l" , "-f" ,
-		"-d" , "-ne" , "-m" ,
-		"-S" , "-v" , "-b" , "-h" } ;
-char	NoResume = 0 ;
-long	Timeout = 300 ;		/* 5 minutes default timeout */
-char	Redial = 0 ;
-char	Silent = 0 ;
-char	NoCheck = 0 ;
-char	Help = 0 ;
-char	ListFile [ MAXLEN ] = { "" } ;
-char	LogFile [ MAXLEN ] = { "" } ;
-char	Directory [ MAXLEN ] = { "./" } ;
-char	NoEstimate = 0 ;
-long	MinRate = 0 ;		/* No minimum rate */
-char	Verbose = 0 ;
-char	StdOut = 0 ;
-char	Identity [ MAXLEN ] = { "linux htget v0.1alpha" } ;
-FILE	* ReportsOutTo ;
+char	ShowHelp ;
+long	Timeout ;
+long	Redial ;
+long	Resume ;
+char	Verbose ;
+long	MinRate ;
+char	ShowBytesRemaining ;
+char	ShowTimeRemaining ;
+char	ShowPercentage ;
+char	ShowBytesPerSecond ;
+char	DownloadsDir [ 128 ] ;
+char	UserAgent [ 128 ] ;
+char	ShowBanner ;
+
+
 
 
 char * MakeHTTPRequest ( char * Hostname , char * Filename ) ;
@@ -105,70 +96,226 @@ void PrintReport ( long ArrivedSize , long RangeSize ,
 
 int main ( int argc , char * argv [] )
 {
-	int		I , Port , J , Success , Ok , AllowResume , Again ;
-	char		RetrieveURL [ MAX_URLS ][ MAXLEN ] ;
-	char		Parameter [ MAXLEN ] ;
-	int		URLs , CurrentURL ;
-	FILE		* ListFileHandle ;
+	char		Temp [ 130 ] ;
+	int			I , J , URLs , Ok , CurrentURL , Success ;
+	FILE		* ResourceFile ;
+	char		* Home ;
+	char		RetrieveURL [ MAX_URLS ][ 128 ] ;
 
+	/* open the resource file. */
+	/* first search cmd line params to see if one has been specified. */
+	for ( I = 1 ; I < argc ; I ++ )
+	{
+		if ( strncasecmp ( "--rcfile=" , argv [ I ] , 9 ) == 0 )
+		{
+			ResourceFile = fopen ( & argv [ I ][ 9 ] , "rt" ) ;
+			if ( ResourceFile == NULL )
+			{
+				fprintf ( stderr , "%s: couldn't open rc file %s\n" , argv [ 0 ] , & argv [ I ][ 9 ] ) ;
+				exit ( 1 ) ;
+			}
+		}
+	}
+	if ( ResourceFile == NULL )
+	{
+		/* try a few other locations */
+		ResourceFile = fopen ( "htgetrc" , "rt" ) ;
+		if ( ResourceFile == NULL )
+		{
+			Home = (char *) getenv ( "HOME" ) ;
+			if ( Home == NULL )
+			{
+				strcpy ( Temp , ".htgetrc" ) ;
+			} else {
+				strcpy ( Temp , Home ) ;
+				strcat ( Temp , "/.htgetrc" ) ;
+			}
+			ResourceFile = fopen ( Temp , "rt" ) ;
+			if ( ResourceFile == NULL)
+			{ 
+				strcpy ( Temp, "/etc/htget.conf" );
+				ResourceFile = fopen ( Temp, "rt" );
+			}
+			if ( ResourceFile == NULL )
+			{
+				fprintf ( stderr , "%s: couldn't find a htget resource file in the current\n"
+						"directory or your home directory. Please read\n"
+						"the documentation that came with htget for more details\n" ,
+						argv [ 0 ] ) ;
+			}
+		}
+	}
+
+	/* set up some default values */
+	Timeout = 30 ;
+	Redial = 10 ;
+	Resume = 1 ;
+	MinRate = 1000 ;
+	Verbose = 0 ;
+	ShowBytesRemaining = 1 ;
+	ShowTimeRemaining = 1 ;
+	ShowPercentage = 1 ;
+	ShowBytesPerSecond = 1 ;
+	strcpy ( DownloadsDir , "./" ) ;
+	strcpy ( UserAgent , "linux htget 0.91" ) ;
+	ShowBanner = 1 ;
+
+	if ( ResourceFile != NULL )
+	{
+		while ( ! feof ( ResourceFile ) )
+		{
+			fgets ( Temp , 128 , ResourceFile ) ;
+			/* Remove terminating newline/junk */
+			for ( I = 0 ; I < strlen ( Temp ) ; I ++ )
+			{
+				if ( ! isprint ( Temp [ I ] ))
+				{
+					Temp [ I ] = '\0' ;
+					break ;
+				}
+			}
+			if ( Temp [ 0 ] != '#' )
+			{
+				if ( strncasecmp ( Temp , "Timeout=" , 8 ) == 0 )
+				{
+					Timeout = atoi ( & Temp [ 8 ] ) ;
+					if ( Timeout < 0 )
+					{
+						fprintf ( stderr , "Timeout value cannot be negative!\n" ) ;
+						return ( 1 ) ;
+					}
+				}
+				if ( strncasecmp ( Temp , "Redial=" , 7 ) == 0 )
+				{
+					Redial = atoi ( & Temp [ 7 ] ) ;
+					if ( Redial < 0 )
+					{
+						fprintf ( stderr , "Redial delay value cannot be negative!\n" ) ;
+						return ( 1 ) ;
+					}
+				}
+				if ( strncasecmp ( Temp , "MinRate=" , 8 ) == 0 )
+				{
+					MinRate = atoi ( & Temp [ 8 ] ) ;
+					if ( MinRate < 0 )
+					{
+						fprintf ( stderr , "MinRate value cannot be negative!\n" ) ;
+						return ( 1 ) ;
+					}
+				}
+				if ( strncasecmp ( Temp , "Resume=" , 7 ) == 0 )
+				{
+					Resume = ( atoi ( & Temp [ 7 ] ) == 1 ) ;
+				}
+				if ( strncasecmp ( Temp , "Verbose=" , 8 ) == 0 )
+				{
+					Verbose = ( atoi ( & Temp [ 8 ] ) == 1 ) ;
+				}
+				if ( strncasecmp ( Temp , "ShowBytesRemaining=" , 19 ) == 0 )
+				{
+					ShowBytesRemaining = ( atoi ( & Temp [ 19 ] ) == 1 ) ;
+				}
+				if ( strncasecmp ( Temp , "ShowPercentage=" , 15 ) == 0 )
+				{
+					ShowPercentage = ( atoi ( & Temp [ 15 ] ) == 1 ) ;
+				}
+				if ( strncasecmp ( Temp , "ShowBytesPerSecond=" , 19 ) == 0 )
+				{
+					ShowBytesPerSecond = ( atoi ( & Temp [ 19 ] ) == 1 ) ;
+				}
+				if ( strncasecmp ( Temp , "ShowBanner=" , 11 ) == 0 )
+				{
+					ShowBanner = ( atoi ( & Temp [ 11 ] ) == 1 ) ;
+				}
+				if ( strncasecmp ( Temp , "DownloadsDir=" , 13 ) == 0 )
+				{
+					strcpy ( DownloadsDir , & Temp [ 13 ] ) ;
+				}
+				if ( strncasecmp ( Temp , "UserAgent=" , 10 ) == 0 )
+				{
+					strcpy ( UserAgent , & Temp [ 10 ] ) ;
+				}
+			}
+		}
+		fclose ( ResourceFile ) ;
+	}
+
+	if ( ShowBanner )
+	{
+		fprintf ( stderr , "HTGET (c) J Whitham 1998  <jwhitham@globalnet.co.uk> version 0.91\n" 
+			"See README. This program comes with NO WARRANTY of any kind.\n\n" ) ;
+	}
+
+	/* parse cmd line */
 	URLs = 0 ;
 	for ( I = 1 ; I < argc ; I ++ )
 	{
 		if ( argv [ I ][ 0 ] == '-' )
 		{
-			for ( J = 0 ; J < OPTIONS ; J ++ )
+			if (( strncasecmp ( argv [ I ] , "--help" , 6 ) == 0 )
+			|| ( strncasecmp ( argv [ I ] , "-h" , 2 ) == 0 )
+			|| ( strncasecmp ( argv [ I ] , "-?" , 2 ) == 0 ))
 			{
-				if ( ( I + 1 ) < argc )
+				ShowHelp = 1 ;
+			}
+			else if ( strncasecmp ( argv [ I ] , "--verbose" , 9 ) == 0 )
+			{
+				Verbose = 1 ;
+			}
+			else if ( strncasecmp ( argv [ I ] , "--noverbose" , 11 ) == 0 )
+			{
+				Verbose = 0 ;
+			}
+			else if ( strncasecmp ( argv [ I ] , "--resume" , 8 ) == 0 )
+			{
+				Resume = 1 ;
+			}
+			else if (( strncasecmp ( argv [ I ] , "--noresume" , 10 ) == 0 )
+			|| ( strncasecmp ( argv [ I ] , "-n" , 2 ) == 0 ))
+			{
+				Resume = 0 ;
+			}
+			else if ( strncasecmp ( argv [ I ] , "--downloadsdir=" , 15 ) == 0 )
+			{
+				strcpy ( DownloadsDir , & argv [ I ][ 15 ] ) ;
+			}
+			else if ( strncasecmp ( argv [ I ] , "--currentdir" , 12 ) == 0 )
+			{
+				strcpy ( DownloadsDir , "./" ) ;
+			}
+			else if ( strncasecmp ( argv [ I ] , "--timeout=" , 10 ) == 0 )
+			{
+				Timeout = atoi ( & argv [ I ][ 10 ] ) ;
+				if ( Timeout < 0 )
 				{
-					strcpy ( Parameter , argv [ I + 1 ] ) ;
-				} else {
-					strcpy ( Parameter , "" ) ;
+					fprintf ( stderr , "Timeout must be a positive integer\n" ) ;
+					return ( 1 ) ;
 				}
-				if (( strcmp ( LongOption [ J ] , argv [ I ] ) == 0 )
-				|| ( strcmp ( ShortOption [ J ] , argv [ I ] ) == 0 ))
+			}
+			else if ( strncasecmp ( argv [ I ] , "--redial=" , 9 ) == 0 )
+			{
+				Redial = atoi ( & argv [ I ][ 9 ] ) ;
+				if ( Redial < 0 )
 				{
-					switch ( J )
-					{
-						case 0 :	NoResume = 1 ;
-								break ;
-						case 1 :	Timeout = 0 ;
-								break ;
-						case 2 :	Timeout = atol ( Parameter ) ;
-								I ++ ;
-								break ;
-						case 3 :	Redial = 1 ;
-								break ;
-						case 4 :	Silent = 1 ;
-								break ;
-						case 5 :	NoCheck = 1 ;
-								break ;
-						case 6 :	strcpy ( ListFile , Parameter ) ;
-								I ++ ;
-								break ;
-						case 7 :	strcpy ( LogFile , Parameter ) ;
-								I ++ ;
-								break ;
-						case 8 :	strcpy ( Directory , Parameter ) ;
-								I ++ ;
-								break ;
-						case 9 :	NoEstimate = 1 ;
-								break ;
-						case 10 :	MinRate = atol ( Parameter ) ;
-								I ++ ;
-								break ;
-						case 11 :	StdOut = 1 ;
-								break ;
-						case 12 :	Verbose = 1 ;
-								break ;
-						case 13 :	strcpy ( Identity , Parameter ) ;
-								I ++ ;
-								break ;
-						case 14 :	Help = 1 ;
-								break ;
-					}
-					break ;
+					fprintf ( stderr , "Redial must be a positive integer\n" ) ;
+					return ( 1 ) ;
 				}
 			}
+			else if ( strncasecmp ( argv [ I ] , "--minrate=" , 10 ) == 0 )
+			{
+				MinRate = atoi ( & argv [ I ][ 10 ] ) ;
+				if ( MinRate < 0 )
+				{
+					fprintf ( stderr , "MinRate must be a positive integer\n" ) ;
+					return ( 1 ) ;
+				}
+			}
+			else if ( strncasecmp ( argv [ I ] , "--rcfile=" , 9 ) == 0 )
+			{
+				/* already dealt with this one */
+			} else {
+				fprintf ( stderr , "Unrecognised parameter: %s\n" , argv [ I ] ) ;
+			}
 		} else {
 			if ( strlen ( argv [ I ] ) > 0 )
 			{
@@ -182,183 +329,78 @@ int main ( int argc , char * argv [] )
 			}
 		}
 	}
-	if ( (( URLs < 1 ) && ( strlen ( ListFile ) < 1 ) ) || Help )  
+	if ( ( URLs < 1 ) || ShowHelp )  
 	{
 		fprintf ( stderr , "Usage: htget <urls..> [options]\n\n"
-			"--help		-h	This text\n"
-			"--noresume	-nr	Do not attempt to resume files.\n"
-			"--notimeout	-nt	Never let the connection time out.\n"
-			"--timeout <t>	-t <t>	Set the timeout to <t> seconds.\n"
-			"--redial	-r	If the connection times out, reconnect and resume\n"
-			"			the download instead of returning to the console.\n" 
-			"--silent	-s	Dont output download info to stderr.\n"
-			"--nocheck	-nc	When resuming files, dont check the file time/dates\n"
-			"			match.\n"
-			"--list <l>	-l <l>	Download files listed in file <l>.\n"
-			"--logfile <l>	-f <l>	Make a log file <l>\n"
-			"--directory <d> -d <d>	Output files to alternate directory <d>\n"
-			"--noestimate	-ne	Dont estimate the download time remaining\n"
-			"--minrate <r>	-m <r>	Set the minimum download rate to <r> bytes/sec.\n"
-			"--stdout	-S	Output file to stdout. Only works with one file.\n"
-			"--verbose	-v	Lots of information including HTTP headers\n"
-			"--be <n>	-b <n>	Set the user-agent to <n>\n\n" ) ;
-		return ( 0 ) ;
+			"Options include:\n"
+			"--help      -h  This text\n"
+			"--verbose       Verbose download mode on\n"
+			"--noverbose     Verbose download mode off\n"
+			"--resume        D/load resuming on\n"
+			"--noresume      D/load resuming off\n"
+			"--timeout=t     Set timeout to t seconds\n"
+			"--rcfile=rc     Specify alternative resource file rc\n"
+			"\n\n"
+			"For more information, see the htget README file\n"
+			"or enter: man htget\n" ) ;
+		return ( 1 ) ;
 	}
-	fprintf ( stderr , BANNER ) ;
 
-	if ( StdOut && ( URLs > 1 ))
-	{
-		fprintf ( stderr , "You can't send multiple files to stdout.\n" ) ;
-		return ( 0 ) ;
-	}
-	if ( strlen ( LogFile ) > 1 )
-	{
-		ReportsOutTo = fopen ( LogFile , "at" ) ;
-		if ( ReportsOutTo == NULL )
-		{
-			fprintf ( stderr , "Couldn't open log file %s\n" , LogFile ) ;
-			return ( 0 ) ;	
-		}
-		fprintf ( ReportsOutTo , "\n" ) ;
-		fprintf ( ReportsOutTo , BANNER ) ;
-		Silent = 0 ;		/* because there is no
-			output to the screen anyway */ 
-		Verbose = 1 ;	
-	} else {
-		ReportsOutTo = stderr ;
-	}
 	Ok = 1 ;
-	AllowResume = 1 ;
 	CurrentURL = 0 ;
 	Success = 0 ;
 	while (( Ok ) && ( CurrentURL < URLs ))
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "\n" ) ;
-		}
+		fprintf ( stderr , "\n" ) ;
 		Ok = 1 ;
-		switch ( HTGet ( RetrieveURL [ CurrentURL ] , AllowResume ) )
+		switch ( HTGet ( RetrieveURL [ CurrentURL ] , Resume ) )
 		{
-			case ERROR_TIMEOUT :	/* just try again */
-						AllowResume = 1 ;
-						break ;
-			case ERROR_WRITE :	if ( ! Silent )
-						{
-							fprintf ( ReportsOutTo , "\nError writing to local file!\n" ) ;
-						}	
+			case ERROR_WRITE :	
+						fprintf ( stderr , "\nError writing to local file!\n" ) ;
 						CurrentURL ++ ;
 						break ;
-			case ERROR_RESUME_IMPOSSIBLE :	/* I hate this! */
-						AllowResume = 0 ;
-						break ;
 			case ERROR_NOT_NEEDED :	
 						CurrentURL ++ ;
 						break ;
-			case ERROR_TCP :	/* All of these are unrecoverable errors */
-			case ERROR_ABORT :	Ok = 0 ;
-						break ;
-			case ERROR_HTTP :	CurrentURL ++ ;
-						break ;
+					/* All of these are unrecoverable errors */
 			case ERROR_PARSE :	CurrentURL ++ ;	/* next please */
 						break ;
 			case DOWNLOAD_OK :	Success ++ ;
 						CurrentURL ++ ;	
 						break ;
+			case ERROR_RESUME_IMPOSSIBLE :	/* I hate this! */
+						break ;
+			case ERROR_TCP :
+			case ERROR_HTTP :
+						if ( Redial == 0 )
+						{
+							CurrentURL ++ ;
+						} else {
+							fprintf ( stderr , "\nRedial: delaying for %d seconds before retry.." , Redial ) ;
+							sleep ( Redial ) ; 
+							fprintf ( stderr , "\n" ) ;
+						}
+						break ;
+			case ERROR_FOUR_OH_FOUR :
+						fprintf ( stderr , "\nFile not found on server.\n" ) ;
+						CurrentURL ++ ;	
+						break ;
+			case ERROR_ABORT :	
+						Ok = 0 ;
+						break ;
+			case ERROR_TIMEOUT :	/* just try again */
+						if ( Redial > 0 )
+						{
+							fprintf ( stderr , "\nRedial: delaying for %d seconds before retry.." , Redial ) ;
+							sleep ( Redial ) ; 
+							fprintf ( stderr , "\n" ) ;
+						}
+						break ;
 		}				
 	}
-	if ( strlen ( ListFile ) > 0 )
-	{
-		ListFileHandle = fopen ( ListFile , "rt" ) ;
-		if ( ListFileHandle == NULL )
-		{
-			if ( ! Silent )
-			{
-				fprintf ( ReportsOutTo , "Couldn't open list file %s\n" , ListFile ) ;
-			} 
-		} else {	 
-			Ok = 1 ;
-			Again = 0 ;
-			while ( Ok )
-			{
-				if ( ! Silent )
-				{
-					fprintf ( ReportsOutTo , "\n" ) ;
-				}
-				if ( ! Again )
-				{
-					if ( feof ( ListFileHandle ) )
-					{
-						Ok = 0 ;
-						I = ERROR_LIST ;
-					} else {
-						fgets ( RetrieveURL [ 0 ] , MAXLEN - 2 , ListFileHandle ) ;
-						for ( I = strlen ( RetrieveURL [ 0 ] ) - 1 ; I >= 0 ; I -- )
-						{
-							if ( ( isalpha ( RetrieveURL [ 0 ][ I ] ))
-							|| ( isdigit ( RetrieveURL [ 0 ][ I ] )) )
-							{
-								break ;
-							} else {	
-								RetrieveURL [ 0 ][ I ] = '\0' ;
-							}
-						}
-						if ( strlen ( RetrieveURL [ 0 ] ) > 0 )
-						{
-							I = HTGet ( RetrieveURL [ 0 ] , AllowResume ) ;
-						} else {
-							I = ERROR_LIST ;
-						}
-					}	
-				} else {
-					I = HTGet ( RetrieveURL [ 0 ] , AllowResume ) ;
-				}
-				Again = 1 ;
-				switch ( I )
-				{
-					case ERROR_TIMEOUT :	/* just try again */
-								AllowResume = 1 ;
-								break ;
-					case ERROR_WRITE :	if ( ! Silent )
-								{
-									fprintf ( ReportsOutTo , "\nError writing to local file!\n" ) ;
-								}	
-								Again = 0 ;
-								break ;
-					case ERROR_RESUME_IMPOSSIBLE :	/* I hate this! */
-								AllowResume = 0 ;
-								break ;
-					case ERROR_NOT_NEEDED :	
-								Again = 0 ;
-								break ;
-					case ERROR_TCP :	/* All of these are unrecoverable errors */
-					case ERROR_ABORT :	Ok = 0 ;
-								break ;
-					case ERROR_HTTP :	Again = 0 ;
-								break ;
-					case ERROR_PARSE :	Again = 0 ;
-								break ;
-					case DOWNLOAD_OK :	Success ++ ;
-								Again = 0 ;
-								break ;
-					case ERROR_LIST :	Ok = 0 ;
-								break ;
-				}				
-			}
-			fclose ( ListFileHandle ) ;
-		}
-	}
-			
-	if ( ! Silent )
-	{
-		fprintf ( ReportsOutTo , "\n\nSuccessfully downloaded %d files\n" ,
-			Success ) ;
-	}		
-	if ( ReportsOutTo != stderr )
-	{
-		fclose ( ReportsOutTo ) ;
-	}
-	return ( 1 ) ;
+	
+	fprintf ( stderr , "\n\nhtget: Successfully downloaded %d of %d URLs\n" , Success , URLs ) ;
+	return ( 0 ) ;
 }
 
 int HTGet ( char * RetrieveURL , int AllowResume ) 
@@ -396,55 +438,44 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 	
 	if ( ! ProcessURL ( RetrieveURL , Hostname , Filename , ActualFilename , & Port ) )
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "Couldn't parse URL %s\n" , RetrieveURL ) ;
-		}	
+		fprintf ( stderr , "Couldn't parse URL %s\n" , RetrieveURL ) ;
 		return ( ERROR_PARSE ) ;
 	}
 
 	/* apply alternate directory stuff */
-	for ( I = strlen ( Directory ) - 1 ; I >= 0 ; I -- )
+	for ( I = strlen ( DownloadsDir ) - 1 ; I >= 0 ; I -- )
 	{
-		if ( Directory [ I ] == '/' )
+		if ( DownloadsDir [ I ] == '/' )
 		{
-			Directory [ I ] == '\0' ;
+			DownloadsDir [ I ] == '\0' ;
 		} else {
 			break ;
 		}
 	}
-	strcat ( Directory , "/" ) ;
+	strcat ( DownloadsDir , "/" ) ;
 	strcpy ( recv_buf , ActualFilename ) ;
-	strcpy ( ActualFilename , Directory ) ;
+	strcpy ( ActualFilename , DownloadsDir ) ;
 	strcat ( ActualFilename , recv_buf ) ;
 	
 	CurrentSize = 0 ;
-	if ( ! StdOut )
+	/* read/resume check */
+	File = fopen ( ActualFilename , "rb" ) ;
+	if ( File != NULL )
 	{
-		/* read/resume check */
-		File = fopen ( ActualFilename , "rb" ) ;
-		if ( File != NULL )
-		{
-			fseek ( File , 0 , SEEK_END ) ;
-			CurrentSize = ftell ( File ) ;
-			fclose ( File ) ;
-		}
-		/* write check */
-		File = fopen ( ActualFilename , "ab" ) ;
-		if ( File == NULL )
-		{
-			if ( ! Silent )
-			{
-				fprintf ( ReportsOutTo , "Error: attempt to open file %s for writing failed.\n" , ActualFilename ) ;
-			}
-			return ( ERROR_WRITE ) ;
-		}
+		fseek ( File , 0 , SEEK_END ) ;
+		CurrentSize = ftell ( File ) ;
 		fclose ( File ) ;
 	}
+	/* write check */
+	File = fopen ( ActualFilename , "ab" ) ;
+	if ( File == NULL )
+	{
+		fprintf ( stderr , "Error: attempt to open file %s for writing failed.\n" , ActualFilename ) ;
+		return ( ERROR_WRITE ) ;
+	}
+	fclose ( File ) ;
 	ActualSize = -1 ;
-	/* NoResume is a command line switch.
-	AllowResume is set false if the resume fails 1st time */
-	if (( CurrentSize > 0 ) && ( ! NoResume ) && ( ! StdOut ) && ( AllowResume ))
+	if (( CurrentSize > 0 ) && AllowResume )
 	{
 		/* See if resume is possible by finding out the file length,
 		time and date */
@@ -454,24 +485,30 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 			return ( ERROR_TCP ) ; /* error messages already generated by TCPConnect */
 		}
 	
-		fprintf ( ReportsOutTo , "Getting header for %s.." , RetrieveURL ) ;
+		fprintf ( stderr , "Getting header for %s.." , RetrieveURL ) ;
 		sprintf ( Request , "HEAD %s HTTP/1.0\r\n"
 			"Host: %s\r\n"
 			"User-agent: %s\r\n"
 			"Authorization: Basic %s\r\n\r\n" ,
-			    Filename , Hostname , Identity ,
+			    Filename , Hostname , UserAgent ,
 			    ToBase64 ( UserPass , strlen ( UserPass ) ) ) ;
 		if ( Verbose )
 		{
-			fprintf ( ReportsOutTo , "\n** Sent: %s" , Request ) ;
+			fprintf ( stderr , "\n** Sent: %s" , Request ) ;
 		}
-		if ( ! GetHeaderForRequest ( socket , Request , Header ) )
+		I = GetHeaderForRequest ( socket , Request , Header ) ;
+		if ( I > 0 )
 		{
-			return ( ERROR_HTTP ) ;
+			if ( I == 404 )
+			{
+				return ( ERROR_FOUR_OH_FOUR ) ;
+			} else {
+				return ( ERROR_HTTP ) ;
+			}
 		}
 		if ( Verbose )
 		{
-			fprintf ( ReportsOutTo , "\n** Received: %s" , Header ) ;
+			fprintf ( stderr , "\n** Received: %s" , Header ) ;
 		}
 		/* get some data from the header..
 		 * the actual size for file resumes */
@@ -484,17 +521,11 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 				if ( CurrentSize < ActualSize )
 				{
 					Resuming = 1 ;
-					if ( ! Silent )
-					{
-						fprintf ( ReportsOutTo , "Resuming download from %ld bytes\n" , CurrentSize ) ;
-					}	
+					fprintf ( stderr , "Resuming download from %ld bytes\n" , CurrentSize ) ;
 				}
 				if ( CurrentSize == ActualSize )
 				{
-					if ( ! Silent )
-					{
-						fprintf ( ReportsOutTo , "Remote and local size is the same!\n" ) ;
-					}	
+					fprintf ( stderr , "Remote and local size is the same!\n" ) ;
 					return ( ERROR_NOT_NEEDED ) ;
 				}
 			}
@@ -511,10 +542,7 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 	{
 		return ( ERROR_TCP ) ;
 	}
-	if ( ! Silent )
-	{
-		fprintf ( ReportsOutTo , "Getting data for %s.." , RetrieveURL ) ;
-	}	
+	fprintf ( stderr , "Getting data for %s.." , RetrieveURL ) ;
 	if ( Resuming )
 	{
 		sprintf ( Request , "GET %s HTTP/1.1\r\n"
@@ -522,7 +550,7 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 			"User-agent: %s\r\n" 
 			"Range: bytes=%ld-%ld\r\n"
 			"Authorization: Basic %s\r\n\r\n" ,
-				Filename , Hostname , Identity ,
+				Filename , Hostname , UserAgent ,
 				CurrentSize , ActualSize ,
 				ToBase64 ( UserPass , strlen ( UserPass ) ) ) ;
 	} else {
@@ -530,21 +558,27 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 			"Host: %s\r\n"
 			"User-agent: %s\r\n" 
 			"Authorization: Basic %s\r\n\r\n" ,
-				Filename , Hostname , Identity ,
+				Filename , Hostname , UserAgent ,
 				ToBase64 ( UserPass , strlen ( UserPass ) ) ) ;
 	}
 	if ( Verbose )
 	{
-		fprintf ( ReportsOutTo , "\n** Sent: %s" , Request ) ;
+		fprintf ( stderr , "\n** Sent: %s" , Request ) ;
 	}
-	if ( ! GetHeaderForRequest ( socket , Request , Header ) )
-	{
-		return ( ERROR_HTTP ) ;
+	I = GetHeaderForRequest ( socket , Request , Header );
+	if ( I > 0 )
+	{	
+		if ( I == 404 )
+		{
+			return ( ERROR_FOUR_OH_FOUR ) ;
+		} else {
+			return ( ERROR_HTTP ) ;
+		}
 	}
 
 	if ( Verbose )
 	{
-		fprintf ( ReportsOutTo , "\n** Received: %s" , Header ) ;
+		fprintf ( stderr , "\n** Received: %s" , Header ) ;
 	}
 	if ( Resuming )
 	{
@@ -571,56 +605,45 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 	if (( Resuming )
 	&& (( RangeStart > CurrentSize ) || ( RangeStart < 128 )) )
 	{
-		fprintf ( ReportsOutTo , "This server does not support resume\n" ) ;
+		fprintf ( stderr , "This server does not support resume\n" ) ;
 		return ( ERROR_RESUME_IMPOSSIBLE ) ;
 	}
 
-	if ( ! StdOut )
+	if ( CurrentSize > 0 )
 	{
-		if ( CurrentSize > 0 )
-		{
-			File = fopen ( ActualFilename , "ab" ) ;
-		} else {
-			File = fopen ( ActualFilename , "wb" ) ;
-		}
-		if ( File == NULL )
-		{
-			return ( ERROR_WRITE ) ;
-		}
+		File = fopen ( ActualFilename , "ab" ) ;
+	} else {
+		File = fopen ( ActualFilename , "wb" ) ;
+	}
+	if ( File == NULL )
+	{
+		return ( ERROR_WRITE ) ;
 	}
 	if ( RangeStart > CurrentSize )
 	{
-		fprintf ( ReportsOutTo , "The HTTP range is past the end of file\n" ) ;
+		fprintf ( stderr , "The HTTP range is past the end of file\n" ) ;
 		return ( ERROR_RESUME_IMPOSSIBLE ) ;
 			/* well actually it isn't but this will force the
 			gizmo to try again from 0 bytes */
 	}
-	if ( ! StdOut )
-	{
-		fseek ( File , RangeStart , SEEK_SET ) ;
-	}
+	fseek ( File , RangeStart , SEEK_SET ) ;
  
  	
-	fprintf ( ReportsOutTo , "\nReceiving HTTP data.." ) ;
+	fprintf ( stderr , "\nReceiving HTTP data.." ) ;
 	I = read ( socket , recv_buf , sizeof ( recv_buf ) ) ;
 	fcntl ( socket , F_SETFL , O_NONBLOCK ) ;
 	
 	ArrivedSize = 0 ;
-	fprintf ( ReportsOutTo , "..\nTotal size: %ld bytes\n" , ActualSize ) ;
+	fprintf ( stderr , "..\nTotal size: %ld bytes\n" , ActualSize ) ;
 	StartTime = time ( NULL ) ;
 	TimeoutTime = StartTime + Timeout ;
 	while ( I > 0 )
 	{	
-		if ( StdOut )
+		if ( fwrite ( recv_buf , I , 1 , File ) != 1 )
 		{
-			fwrite ( recv_buf , I , 1 , stdout ) ;
-		} else {
-			if ( fwrite ( recv_buf , I , 1 , File ) != 1 )
-			{
-				fprintf ( ReportsOutTo , "\nError writing to file %s\n" , ActualFilename ) ;
-				fclose ( File ) ;
-				return ( ERROR_WRITE ) ;
-			}
+			fprintf ( stderr , "\nError writing to file %s\n" , ActualFilename ) ;
+			fclose ( File ) ;
+			return ( ERROR_WRITE ) ;
 		}
 		ArrivedSize += I ;
 		CurrentTime = time ( NULL ) ;
@@ -647,9 +670,7 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 		while ( I < 0 )
 		{
 			I = read ( socket , recv_buf , sizeof ( recv_buf ) ) ;
-			if (( ! Silent )
-			&& ( ReportsOutTo == stderr )
-			&& ( RedundantReportsTimer != CurrentTime ))
+			if ( RedundantReportsTimer != CurrentTime )
 			{
 				PrintReport ( ArrivedSize , RangeSize , CurrentTime , StartTime , TimeoutTime ) ;
 				RedundantReportsTimer = CurrentTime ;
@@ -661,22 +682,13 @@ int HTGet ( char * RetrieveURL , int AllowResume )
 			}
 			if ( CurrentTime > TimeoutTime )
 			{
-				if ( ! StdOut )
-				{
-					fclose ( File ) ;
-				}
+				fclose ( File ) ;
 				return ( ERROR_TIMEOUT ) ;
 			}
 		}
 	}
-	if ( ! StdOut )
-	{
-		fclose ( File ) ;
-	}
-	if ( ! Silent )
-	{
-		fprintf ( ReportsOutTo , "\nFinished downloading.\n" ) ;
-	}
+	fclose ( File ) ;
+	fprintf ( stderr , "\nFinished downloading.\n" ) ;
 	return ( DOWNLOAD_OK ) ;
 }
 
@@ -718,20 +730,14 @@ int TCPConnect ( char * Hostname , unsigned int Port )
 
 	if( ( host = gethostbyname ( Hostname ) ) == NULL ) 
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "Error finding host %s\n" , Hostname ) ;
-		}
+		fprintf ( stderr , "Error finding host %s\n" , Hostname ) ;
 		return ( 0 ) ;
 	}
 
 /* get the socket */
 	if( ( sock_fd = socket ( AF_INET , SOCK_STREAM , 0 ) ) < 0 )
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "Error creating TCP/IP socket\n" ) ;
-		}
+		fprintf ( stderr , "Error creating TCP/IP socket\n" ) ;
 		return ( 0 ) ;
 	}
 
@@ -742,10 +748,7 @@ int TCPConnect ( char * Hostname , unsigned int Port )
   
 	if ( connect ( sock_fd , ( struct sockaddr * ) & sin , sizeof ( sin ) ) < 0 )
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "Error connecting to %s\n" , Hostname ) ;
-		}
+		fprintf ( stderr , "Error connecting to %s\n" , Hostname ) ;
 		return ( 0 ) ;
 	}
 
@@ -869,7 +872,8 @@ int ProcessURL ( char * TheURL , char * Hostname , char * Filename , char * Actu
 	if ( strlen ( BufferURL ) < 1 )
 	{
 		/* no filename error! */
-		return ( 0 ) ;
+		/* Bugfix: this is still legal, but we must specify root */
+		strcat ( BufferURL , "/" ) ;
 	}
 	(* Port) = 80 ; /* default HTTP port */
 	for ( I = strlen ( BufferURL ) - 1 ; I >= 0 ; I -- )
@@ -900,10 +904,11 @@ int ProcessURL ( char * TheURL , char * Hostname , char * Filename , char * Actu
 	}
 	if ( strlen ( ActualFilename ) < 1 )
 	{
-		return ( 0 ) ; 
-		/* because that means this is a directory
-		download. And I can't do them. */
+		/* This is probably because user asked for URL like: www.stuff.com/
+		Make up a filename based on URL. */
+		strcpy ( ActualFilename , Hostname ) ;
 	}
+
 	return ( 1 ) ;
 }
 
@@ -930,44 +935,29 @@ int GetHeaderForRequest ( unsigned socket , char * request , char * Header )
 	but I put it in after some problems connecting to my local server,
 	there are possibly some latency issues here? */
 	
-	if ( ! Silent )
-	{
-		fprintf ( ReportsOutTo , ".." ) ;
-	}
+	fprintf ( stderr , ".." ) ;
 	ch = '\0' ;
 	I = read ( socket , & ch , 1 ) ;
 	while ( I == 0 )
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "." ) ;
-		}
+		fprintf ( stderr , "." ) ;
 		sleep ( 1 ) ;	
 		I = read ( socket , & ch , 1 ) ;
 	}
 
 	if ( I == -1 )
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "\nUnknown TCP/IP error\n" ) ;
-		}
-		return ( 0 ) ;
+		fprintf ( stderr , "\nUnknown TCP/IP error\n" ) ;
+		return ( 1 ) ;
 	}
 
 	strcpy ( status , "" ) ;
 	if ( ! ReadLine ( socket , status ) )
 	{
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "no response\n" )  ;
-		}
-		return ( 0 ) ;
+		fprintf ( stderr , "no response\n" )  ;
+		return ( 1 ) ;
 	} else {
-		if ( ! Silent )
-		{
-			fprintf ( ReportsOutTo , "\n" ) ;
-		}
+		fprintf ( stderr , "\n" ) ;
 	}
 	/* this will re-append the check character read above to the 1st line */
 	line [ 0 ] = ch ;
@@ -991,11 +981,13 @@ int GetHeaderForRequest ( unsigned socket , char * request , char * Header )
 		strcpy(token, (char *)strtok(NULL, " "));
 		if ( token [ 0 ] == '4' || token [ 0 ] == '5' )
 		{
-			if ( ! Silent )
+			fprintf ( stderr , "http error from server: %s" , status ) ;
+			if ( strncmp ( token , "404" , 3 ) == 0 )
 			{
-				fprintf ( ReportsOutTo , "http error from server: %s" , status ) ;
+				return ( 404 ) ;
+			} else {
+				return ( 1 ) ;
 			}
-			return(0);
 		}
 	}
 		
@@ -1014,7 +1006,7 @@ int GetHeaderForRequest ( unsigned socket , char * request , char * Header )
 		}
 		strcat ( Header , line ) ;
 	}
-	return ( 1 ) ;
+	return ( 0 ) ;
 }
 
 
@@ -1065,18 +1057,29 @@ void PrintReport ( long ArrivedSize , long RangeSize ,
 	time_t		TimeLeft ;
 	long		BytesLeft ;
 
-	if ( RangeSize > 0 )
+	fprintf ( stderr , "\r\033[K\r.." ) ;
+	if ( ShowBytesRemaining )
+	{
+		if ( RangeSize > 0 )
+		{
+			fprintf ( stderr , " %ld/%ld" , ArrivedSize , RangeSize ) ;
+		} else {
+			fprintf ( stderr , " %ld/unknown" , ArrivedSize ) ;
+		}
+	}
+	if (( ShowPercentage ) && ( RangeSize > 0 ) )
 	{
 		Percentage = ( ArrivedSize * 100 ) / RangeSize ;
-		fprintf ( stderr , "\r\033[K\r%ld/%ld (%d%%)" , ArrivedSize , RangeSize , Percentage ) ;
-	} else {
-		fprintf ( stderr , "\r\033[K\r%ld/unknown" , ArrivedSize ) ;
+		fprintf ( stderr , " %d%%" , Percentage ) ;
 	}
 	if ( ( CurrentTime - StartTime ) > 0 )
 	{
 		BytesPerSecond = ArrivedSize / ( CurrentTime - StartTime ) ;
-		fprintf ( stderr , " %ld bytes/sec" , BytesPerSecond ) ;
-		if (( RangeSize > 0 ) && ! NoEstimate )
+		if ( ShowBytesPerSecond )
+		{
+			fprintf ( stderr , " %ld bytes/sec" , BytesPerSecond ) ;
+		}
+		if (( RangeSize > 0 ) && ShowTimeRemaining )
 		{
 			/* time left may be predicted. */
 			BytesLeft = RangeSize - ArrivedSize ;
